import dayjs from 'dayjs';
import validator from 'validator';
import { ulid } from 'ulid';

class ValueObject {
    _input;
    _setInner = false;
    _inner;
    _setOutput = false;
    _output;
    constructor(_input) {
        this._input = _input;
    }
    fromInput() {
        return this.input;
    }
    toOutput() {
        return this.inner;
    }
    get input() {
        return this._input;
    }
    get inner() {
        if (!this._setInner) {
            this._setInner = true;
            this._inner = this.fromInput();
        }
        // biome-ignore lint/style/noNonNullAssertion:
        return this._inner;
    }
    get value() {
        if (!this._setOutput) {
            this._setOutput = true;
            this._output = this.toOutput();
            Object.freeze(this._output);
        }
        // biome-ignore lint/style/noNonNullAssertion:
        return this._output;
    }
    equals(value) {
        return this.compare(value) === 0;
    }
}

const compareNullable = (a, b) => {
    if (a === null) {
        if (b === null) {
            return 0;
        }
        return 1;
    }
    return -1;
};

// Inner type = string: consider serialization
class DateObject extends ValueObject {
    fromInput() {
        const input = this.input;
        if (input === null) {
            return null;
        }
        return dayjs(input).toISOString();
    }
    toOutput() {
        if (this.inner === null) {
            return null;
        }
        return dayjs(this.inner);
    }
    getErrors(name) {
        if (typeof this.input === 'string' &&
            !(validator.isDate(this.input) || validator.isISO8601(this.input))) {
            return [{ name, error: '日付の形式が正しくありません' }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        if (this.value.isSame(value.value)) {
            return 0;
        }
        if (this.value.isBefore(value.value)) {
            return -1;
        }
        return 1;
    }
}

class Flags extends ValueObject {
    getErrors(name) {
        if (this.input === null) {
            return undefined;
        }
        if (!this.flagTypes.includes(this.input)) {
            return [{ name, error: `定義されていないフラグです: ${this.input}` }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        return this.value.localeCompare(value.value);
    }
}

class Float extends ValueObject {
    fromInput() {
        const input = this.input;
        if (input === null) {
            return null;
        }
        if (typeof input === 'number') {
            return input;
        }
        let num = Number.parseFloat(input);
        const max = this.getMaxNumber();
        if (max !== undefined && num > max && this.isTruncateMode()) {
            num = Math.min(max, num);
        }
        const min = this.getMinNumber();
        if (min !== undefined && num < min && this.isTruncateMode()) {
            num = Math.max(min, num);
        }
        return num;
    }
    getMaxNumber() {
        return undefined;
    }
    getMinNumber() {
        return undefined;
    }
    isTruncateMode() {
        return false;
    }
    getErrors(name) {
        const num = this.fromInput();
        if (num === null) {
            return undefined;
        }
        if (typeof this.input === 'string' && !validator.isNumeric(this.input)) {
            return [{ name, error: '数値の形式が正しくありません' }];
        }
        const max = this.getMaxNumber();
        if (max !== undefined && num > max) {
            return [{ name, error: `${max}以下の値を入力してください` }];
        }
        const min = this.getMinNumber();
        if (min !== undefined && num < min) {
            return [{ name, error: `${min}以上の値を入力してください` }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        const diff = this.value - value.value;
        if (diff < 0)
            return -1;
        if (diff > 0)
            return 1;
        return 0;
    }
}

class Int extends Float {
    fromInput() {
        const input = super.fromInput();
        if (input === null) {
            return null;
        }
        return Math.floor(input);
    }
    getErrors(name) {
        if (this.input === null) {
            return undefined;
        }
        const results = super.getErrors(name);
        if (results?.length) {
            return results;
        }
        if (typeof this.input === 'string' && !validator.isInt(this.input)) {
            return [{ name, error: '整数の形式が正しくありません' }];
        }
        const num = this.fromInput();
        if (!Number.isSafeInteger(num)) {
            return [{ name, error: '有効な整数ではありません' }];
        }
        return undefined;
    }
}

class StringId extends ValueObject {
    fromInput() {
        if (this.input === undefined) {
            return ulid();
        }
        if (this.input === null) {
            return null;
        }
        return `${this.input}`;
    }
    getErrors(name) {
        const text = this.fromInput();
        if (text === null) {
            return undefined;
        }
        if (!text.length) {
            return [{ name, error: '値を指定してください' }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        return this.value.localeCompare(value.value);
    }
}

class Text extends ValueObject {
    fromInput() {
        const input = this.input;
        if (input === null) {
            return null;
        }
        return `${input}`;
    }
    getValidationMinLength() {
        return undefined;
    }
    getValidationMaxLength() {
        return undefined;
    }
    getErrors(name) {
        const text = this.fromInput();
        if (text === null) {
            return undefined;
        }
        const results = [];
        if (!text.length) {
            results.push({ name, error: '値を指定してください' });
        }
        else {
            const min = this.getValidationMinLength();
            if (min && text.length < min) {
                results.push({ name, error: `${min}文字より長く入力してください` });
            }
        }
        const max = this.getValidationMaxLength();
        if (max && text.length > max) {
            results.push({ name, error: `${max}文字より短く入力してください` });
        }
        return results.length ? results : undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        return this.value.localeCompare(value.value);
    }
}

class Url extends ValueObject {
    getErrors(name) {
        if (this.input === null) {
            return undefined;
        }
        if (!validator.isURL(this.input)) {
            return [{ name, error: 'URLの形式が正しくありません' }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value === null || value.value === null) {
            return compareNullable(this.value, value.value);
        }
        return this.value.localeCompare(value.value);
    }
}

class Exception extends Error {
    status;
    context;
    constructor(status, message, 
    // biome-ignore lint/suspicious/noExplicitAny:
    context) {
        super(message);
        this.status = status;
        this.context = context;
    }
}

class InvalidUsage extends Exception {
    constructor(reason) {
        super(500, '使用方法が正しくありません', { reason });
    }
}

class ValidationException extends Exception {
    errors;
    constructor(errors) {
        super(422, 'バリデーションエラーが発生しました', errors);
        this.errors = errors;
    }
}

class Collection {
    collections;
    constructor(collections) {
        this.collections = collections;
        Object.freeze(this.collections);
    }
    [Symbol.iterator]() {
        return this.collections.values();
    }
    find(filter) {
        return this.collections.find(filter);
    }
    filter(filter) {
        return this.collections.filter(filter);
    }
    map(callback) {
        return this.collections.map(callback);
    }
    sorted(callback) {
        return this.collections.slice().sort(callback);
    }
    isEmpty() {
        return !this.collections.length;
    }
    count() {
        return this.collections.length;
    }
    getErrors(prev) {
        const getPrev = (item) => (prev?.collections ?? []).find((x) => item.equals(x));
        const errors = this.collections.reduce((acc, item, index) => ({
            // biome-ignore lint/performance/noAccumulatingSpread:
            ...acc,
            ...Object.fromEntries(Object.entries(item.getErrors(getPrev(item))).map(([key, value]) => [
                `${key}[${index}]`,
                value,
            ])),
        }), {});
        if (Object.keys(errors).length) {
            return errors;
        }
        return undefined;
    }
}

// biome-ignore lint/suspicious/noExplicitAny:
class Entity {
    static _isCreating = false;
    constructor() {
        if (!Entity._isCreating) {
            throw new InvalidUsage();
        }
    }
    // biome-ignore lint/suspicious/noExplicitAny:
    static _create(...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            const instance = new this(...args);
            instance.validate();
            return instance;
        }
        finally {
            Entity._isCreating = false;
        }
    }
    static _reconstruct(...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            return new this(...args);
        }
        finally {
            Entity._isCreating = false;
        }
    }
    // biome-ignore lint/suspicious/noExplicitAny:
    static _update(target, ...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            const instance = new this(...args);
            instance.validate(target);
            return instance;
        }
        finally {
            Entity._isCreating = false;
        }
    }
    getErrors(prev) {
        return Object.keys(this).reduce((acc, key) => {
            // biome-ignore lint/suspicious/noExplicitAny:
            const member = this[key];
            const prevValue = prev
                ? // biome-ignore lint/suspicious/noExplicitAny:
                    (prev[key] ?? undefined)
                : undefined;
            if (member && member instanceof Collection) {
                const errors = member.getErrors(prevValue);
                if (errors) {
                    return Object.entries(errors).reduce((acc, [key, value]) => {
                        acc[key] = [...new Set([...(acc[key] ?? []), ...value])];
                        return acc;
                    }, acc);
                }
            }
            if (member && member instanceof ValueObject) {
                const name = key.replace(/^_/, '');
                const errors = member.getErrors(name, prevValue);
                if (errors?.length) {
                    return errors.reduce((acc, error) => {
                        acc[error.name] = [
                            ...new Set([...(acc[error.name] ?? []), error.error]),
                        ];
                        return acc;
                    }, acc);
                }
            }
            return acc;
        }, {});
    }
    // biome-ignore lint/suspicious/noConfusingVoidType:
    validate(prev) {
        const errors = this.getErrors(prev);
        if (Object.keys(errors).length) {
            throw new ValidationException(errors);
        }
    }
}

export { Collection, DateObject, Entity, Flags, Float, Int, StringId, Text, Url, ValueObject };
